@@ -86,6 +86,37 @@ namespace NpgsqlTypes
 
         private static readonly Version Npgsql207 = new Version("2.0.7");
 
+        private static readonly Dictionary<string, NpgsqlBackendTypeInfo> DefaultBackendInfoMapping = PrepareDefaultBackendInfoMapping();
+
+
+        
+        private static Dictionary<string, NpgsqlBackendTypeInfo> PrepareDefaultBackendInfoMapping()
+        {
+            Dictionary<string, NpgsqlBackendTypeInfo> NameIndex = new Dictionary<string, NpgsqlBackendTypeInfo>();
+
+            foreach (NpgsqlBackendTypeInfo TypeInfo in TypeInfoList(false, new Version("10.0.0.0")))
+			{
+				NameIndex.Add(TypeInfo.Name, TypeInfo);
+				
+				//do the same for the equivalent array type.
+                NameIndex.Add("_" + TypeInfo.Name, ArrayTypeInfo(TypeInfo));
+				
+			}
+             
+            return NameIndex;
+        } 
+
+
+        /// <summary>
+        /// Find a NpgsqlNativeTypeInfo in the default types map that can handle objects
+        /// of the given NpgsqlDbType.
+        /// </summary>
+        public static bool TryGetBackendTypeInfo(String BackendTypeName, out NpgsqlBackendTypeInfo TypeInfo)
+        {
+            return DefaultBackendInfoMapping.TryGetValue(BackendTypeName, out TypeInfo);
+                
+        }
+
 
 		/// <summary>
 		/// Find a NpgsqlNativeTypeInfo in the default types map that can handle objects
@@ -253,9 +284,6 @@ namespace NpgsqlTypes
 			NpgsqlNativeTypeMapping nativeTypeMapping = new NpgsqlNativeTypeMapping();
 
 
-
-
-
             nativeTypeMapping.AddType("name", NpgsqlDbType.Name, DbType.String, true, null);
 
 			nativeTypeMapping.AddType("oidvector", NpgsqlDbType.Oidvector, DbType.String, true, null);
@@ -300,7 +328,9 @@ namespace NpgsqlTypes
 
 			nativeTypeMapping.AddType("int2", NpgsqlDbType.Smallint, DbType.Int16, false, BasicNativeToBackendTypeConverter.ToBasicType<short>);
 
-			nativeTypeMapping.AddTypeAlias("int2", typeof (Int16));
+            nativeTypeMapping.AddTypeAlias("int2", typeof (UInt16));
+            
+		    nativeTypeMapping.AddTypeAlias("int2", typeof (Int16));
 
 			nativeTypeMapping.AddDbTypeAlias("int2", DbType.Byte);
 
@@ -344,19 +374,21 @@ namespace NpgsqlTypes
 
 			nativeTypeMapping.AddTypeAlias("time", typeof (NpgsqlTime));
 
-			nativeTypeMapping.AddType("timestamp", NpgsqlDbType.Timestamp, DbType.DateTime, true,
-			                          new ConvertNativeToBackendHandler(ExtendedNativeToBackendTypeConverter.ToTimeStamp));
+            nativeTypeMapping.AddType("timestamptz", NpgsqlDbType.TimestampTZ, DbType.DateTime, true,
+                                      new ConvertNativeToBackendHandler(ExtendedNativeToBackendTypeConverter.ToTimeStamp));
+
+            nativeTypeMapping.AddTypeAlias("timestamptz", typeof(NpgsqlTimeStampTZ));
 
             nativeTypeMapping.AddType("abstime", NpgsqlDbType.Abstime, DbType.DateTime, true,
                                       new ConvertNativeToBackendHandler(ExtendedNativeToBackendTypeConverter.ToTimeStamp));
 
+			nativeTypeMapping.AddType("timestamp", NpgsqlDbType.Timestamp, DbType.DateTime, true,
+			                          new ConvertNativeToBackendHandler(ExtendedNativeToBackendTypeConverter.ToTimeStamp));
+            
 			nativeTypeMapping.AddTypeAlias("timestamp", typeof (DateTime));
 			nativeTypeMapping.AddTypeAlias("timestamp", typeof (NpgsqlTimeStamp));
 
-			nativeTypeMapping.AddType("timestamptz", NpgsqlDbType.TimestampTZ, DbType.DateTime, true,
-			                          new ConvertNativeToBackendHandler(ExtendedNativeToBackendTypeConverter.ToTimeStamp));
-
-			nativeTypeMapping.AddTypeAlias("timestamptz", typeof (NpgsqlTimeStampTZ));
+			
 
 			nativeTypeMapping.AddType("point", NpgsqlDbType.Point, DbType.Object, true,
 			                          new ConvertNativeToBackendHandler(ExtendedNativeToBackendTypeConverter.ToPoint));
@@ -413,7 +445,6 @@ namespace NpgsqlTypes
 
 		private static IEnumerable<NpgsqlBackendTypeInfo> TypeInfoList(bool useExtendedTypes, Version compat)
 		{
-
             yield return new NpgsqlBackendTypeInfo(0, "oidvector", NpgsqlDbType.Text, DbType.String, typeof (String), null);
 
 			yield return new NpgsqlBackendTypeInfo(0, "unknown", NpgsqlDbType.Text, DbType.String, typeof (String), null);
@@ -537,36 +568,36 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
                     yield return
                         new NpgsqlBackendTypeInfo(0, "interval", NpgsqlDbType.Interval, DbType.Object, typeof(NpgsqlInterval),
                                                   new ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToInterval),
-                                                  typeof(TimeSpan), interval => (TimeSpan)(NpgsqlInterval)interval);
+                                                  typeof(TimeSpan), interval => (TimeSpan)(NpgsqlInterval)interval, intervalNpgsql => (intervalNpgsql is TimeSpan ? (NpgsqlInterval)(TimeSpan) intervalNpgsql : intervalNpgsql));
                 }
 
 				yield return
 					new NpgsqlBackendTypeInfo(0, "date", NpgsqlDbType.Date, DbType.Date, typeof (NpgsqlDate),
 					                          new ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToDate),
-                                              typeof(DateTime), date => (DateTime)(NpgsqlDate)date);
+                                              typeof(DateTime), date => (DateTime)(NpgsqlDate)date, npgsqlDate => (npgsqlDate is DateTime ? (NpgsqlDate)(DateTime) npgsqlDate : npgsqlDate));
 
 				yield return
 					new NpgsqlBackendTypeInfo(0, "time", NpgsqlDbType.Time, DbType.Time, typeof (NpgsqlTime),
                                               new ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToTime),
-                                              typeof(DateTime), time => (DateTime)(NpgsqlTime)time);
+                                              typeof(DateTime), time => (DateTime)(NpgsqlTime)time, npgsqlTime => (npgsqlTime is TimeSpan ? (NpgsqlTime)(TimeSpan) npgsqlTime : npgsqlTime));
 
 				yield return
 					new NpgsqlBackendTypeInfo(0, "timetz", NpgsqlDbType.TimeTZ, DbType.Time, typeof (NpgsqlTimeTZ),
                                               new ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToTimeTZ),
-                                              typeof(DateTime), timetz => (DateTime)(NpgsqlTimeTZ)timetz);
+                                              typeof(DateTime), timetz => (DateTime)(NpgsqlTimeTZ)timetz, npgsqlTimetz => (npgsqlTimetz is TimeSpan ? (NpgsqlTimeTZ)(TimeSpan) npgsqlTimetz : npgsqlTimetz));
 
 				yield return
 					new NpgsqlBackendTypeInfo(0, "timestamp", NpgsqlDbType.Timestamp, DbType.DateTime, typeof (NpgsqlTimeStamp),
                                               new ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToTimeStamp),
-                                              typeof(DateTime), timestamp => (DateTime)(NpgsqlTimeStamp)timestamp);
+                                              typeof(DateTime), timestamp => (DateTime)(NpgsqlTimeStamp)timestamp, npgsqlTimestamp => (npgsqlTimestamp is DateTime ? (NpgsqlTimeStamp)(DateTime) npgsqlTimestamp : npgsqlTimestamp));
                 yield return
                     new NpgsqlBackendTypeInfo(0, "abstime", NpgsqlDbType.Abstime, DbType.DateTime, typeof(NpgsqlTimeStampTZ),
                               new ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToTimeStampTZ),
-                              typeof(DateTime), timestamp => (DateTime)(NpgsqlTimeStampTZ)timestamp);
+                              typeof(DateTime), timestamp => (DateTime)(NpgsqlTimeStampTZ)timestamp, npgsqlTimestampTZ => (npgsqlTimestampTZ is DateTime ? (NpgsqlTimeStampTZ)(DateTime) npgsqlTimestampTZ : npgsqlTimestampTZ));
 				yield return
 					new NpgsqlBackendTypeInfo(0, "timestamptz", NpgsqlDbType.TimestampTZ, DbType.DateTime, typeof (NpgsqlTimeStampTZ),
                                               new ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToTimeStampTZ),
-                                              typeof(DateTime), timestamptz => ((DateTime)(NpgsqlTimeStampTZ)timestamptz).ToLocalTime());
+                                              typeof(DateTime), timestamptz => ((DateTime)(NpgsqlTimeStampTZ)timestamptz).ToLocalTime(), npgsqlTimestampTZ => (npgsqlTimestampTZ is DateTime ? (NpgsqlTimeStampTZ)(DateTime) npgsqlTimestampTZ : npgsqlTimestampTZ));
 			}
 		}
 
@@ -704,10 +735,12 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
 	/// <summary>
 	/// Delegate called to convert the given native data to its backand representation.
 	/// </summary>
-	internal delegate String ConvertNativeToBackendHandler(NpgsqlNativeTypeInfo TypeInfo, Object NativeData);
+	internal delegate String ConvertNativeToBackendHandler(NpgsqlNativeTypeInfo TypeInfo, Object NativeData, Boolean ForExtendedQuery);
 
     internal delegate object ConvertProviderTypeToFrameworkTypeHander(object value);
 
+    internal delegate object ConvertFrameworkTypeToProviderTypeHander(object value);
+
 	/// <summary>
 	/// Represents a backend data type.
 	/// This class can be called upon to convert a backend field representation to a native object.
@@ -716,6 +749,7 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
 	{
 		private readonly ConvertBackendToNativeHandler _ConvertBackendToNative;
         private readonly ConvertProviderTypeToFrameworkTypeHander _convertProviderToFramework;
+        private readonly ConvertFrameworkTypeToProviderTypeHander _convertFrameworkToProvider;
 
 		internal Int32 _OID;
 		private readonly String _Name;
@@ -749,11 +783,13 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
 
         public NpgsqlBackendTypeInfo(Int32 OID, String Name, NpgsqlDbType NpgsqlDbType, DbType DbType, Type Type,
                                      ConvertBackendToNativeHandler ConvertBackendToNative, Type frameworkType,
-                                     ConvertProviderTypeToFrameworkTypeHander convertProviderToFramework)
+                                     ConvertProviderTypeToFrameworkTypeHander convertProviderToFramework,
+                                     ConvertFrameworkTypeToProviderTypeHander convertFrameworkToProvider)
             : this(OID, Name, NpgsqlDbType, DbType, Type, ConvertBackendToNative)
         {
             _frameworkType = frameworkType;
             _convertProviderToFramework = convertProviderToFramework;
+            _convertFrameworkToProvider = convertFrameworkToProvider;
         }
 
 		/// <summary>
@@ -852,8 +888,25 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
             }
             return providerValue;
         }
+
+        internal object ConvertToProviderType(object frameworkValue)
+        {
+            if (frameworkValue == DBNull.Value)
+            {
+                return frameworkValue;
+            }
+            else if (_convertFrameworkToProvider!= null)
+            {
+                return _convertFrameworkToProvider(frameworkValue);
+            }
+            
+            return frameworkValue;
+        }
+    
     }
 
+    
+
 	/// <summary>
 	/// Represents a backend data type.
 	/// This class can be called upon to convert a native object to its backend field representation,
@@ -1022,7 +1075,7 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
 			if (_ConvertNativeToBackend != null)
 			{
 				return
-					(this.Quote ? QuoteString(_ConvertNativeToBackend(this, NativeData)) : _ConvertNativeToBackend(this, NativeData));
+					(this.Quote ? QuoteString(_ConvertNativeToBackend(this, NativeData, false)) : _ConvertNativeToBackend(this, NativeData, false));
 			}
 			else
 			{
@@ -1061,7 +1114,7 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
 
 			if (_ConvertNativeToBackend != null)
 			{
-				return _ConvertNativeToBackend(this, NativeData);
+				return _ConvertNativeToBackend(this, NativeData, true);
 			}
 			else
 			{
@@ -1347,4 +1400,185 @@ ConvertBackendToNativeHandler(ExtendedBackendToNativeTypeConverter.ToGuid));
 			return TypeIndex.ContainsKey(Type);
 		}
 	}
+    
+    internal static class ExpectedTypeConverter
+    {
+        internal static object ChangeType(object value, Type expectedType)
+        {
+            if (value == null)
+                return null;
+            Type currentType = value.GetType();
+            if (value is DBNull || currentType == expectedType)
+                return value;
+#if NET35
+            if (expectedType == typeof(DateTimeOffset))
+            {
+                if (currentType == typeof(NpgsqlDate))
+                {
+                    return new DateTimeOffset((DateTime)(NpgsqlDate)value);
+                }
+                else if (currentType == typeof(NpgsqlTime))
+                {
+                    return new DateTimeOffset((DateTime)(NpgsqlTime)value);
+                }
+                else if (currentType == typeof(NpgsqlTimeTZ))
+                {
+                    NpgsqlTimeTZ timetz = (NpgsqlTimeTZ)value;
+                    return new DateTimeOffset(timetz.Ticks, new TimeSpan(timetz.TimeZone.Hours, timetz.TimeZone.Minutes, timetz.TimeZone.Seconds));
+                }
+                else if (currentType == typeof(NpgsqlTimeStamp))
+                {
+                    return new DateTimeOffset((DateTime)(NpgsqlTimeStamp)value);
+                }
+                else if (currentType == typeof(NpgsqlTimeStampTZ))
+                {
+                    NpgsqlTimeStampTZ timestamptz = (NpgsqlTimeStampTZ)value;
+                    return new DateTimeOffset(timestamptz.Ticks, new TimeSpan(timestamptz.TimeZone.Hours, timestamptz.TimeZone.Minutes, timestamptz.TimeZone.Seconds));
+                }
+                else if (currentType == typeof(NpgsqlInterval))
+                {
+                    return new DateTimeOffset(((TimeSpan)(NpgsqlInterval)value).Ticks, TimeSpan.FromSeconds(0));
+                }
+                else if (currentType == typeof(DateTime))
+                {
+                    return new DateTimeOffset((DateTime)value);
+                }
+                else if (currentType == typeof(TimeSpan))
+                {
+                    return new DateTimeOffset(((TimeSpan)value).Ticks, TimeSpan.FromSeconds(0));
+                }
+                else
+                {
+                    return DateTimeOffset.Parse(value.ToString(), CultureInfo.InvariantCulture);
+                }
+            }
+            else
+#endif
+            if (expectedType == typeof(TimeSpan))
+            {
+                if (currentType == typeof(NpgsqlDate))
+                {
+                    return new TimeSpan(((DateTime)(NpgsqlDate)value).Ticks);
+                }
+                else if (currentType == typeof(NpgsqlTime))
+                {
+                    return new TimeSpan(((NpgsqlTime)value).Ticks);
+                }
+                else if (currentType == typeof(NpgsqlTimeTZ))
+                {
+                    return new TimeSpan(((NpgsqlTimeTZ)value).UTCTime.Ticks);
+                }
+                else if (currentType == typeof(NpgsqlTimeStamp))
+                {
+                    return new TimeSpan(((NpgsqlTimeStamp)value).Ticks);
+                }
+                else if (currentType == typeof(NpgsqlTimeStampTZ))
+                {
+                    return new TimeSpan(((DateTime)(NpgsqlTimeStampTZ)value).ToUniversalTime().Ticks);
+                }
+                else if (currentType == typeof(NpgsqlInterval))
+                {
+                    return (TimeSpan)(NpgsqlInterval)value;
+                }
+                else if (currentType == typeof(DateTime))
+                {
+                    return new TimeSpan(((DateTime)value).ToUniversalTime().Ticks);
+                }
+                else if (currentType == typeof(DateTimeOffset))
+                {
+                    return new TimeSpan(((DateTimeOffset)value).Ticks);
+                }
+                else
+                {
+#if NET40
+                    return TimeSpan.Parse(value.ToString(), CultureInfo.InvariantCulture);
+#else
+                    return TimeSpan.Parse(value.ToString());
+#endif
+                }
+            }
+            else if (expectedType == typeof(string))
+            {
+                return value.ToString();
+            }
+            else if (expectedType == typeof(Guid))
+            {
+                if (currentType == typeof(byte[]))
+                {
+                    return new Guid((byte[])value);
+                }
+                else
+                {
+                    return new Guid(value.ToString());
+                }
+            }
+            else if (expectedType == typeof(DateTime))
+            {
+                if (currentType == typeof(NpgsqlDate))
+                {
+                    return (DateTime)(NpgsqlDate)value;
+                }
+                else if (currentType == typeof(NpgsqlTime))
+                {
+                    return (DateTime)(NpgsqlTime)value;
+                }
+                else if (currentType == typeof(NpgsqlTimeTZ))
+                {
+                    return (DateTime)(NpgsqlTimeTZ)value;
+                }
+                else if (currentType == typeof(NpgsqlTimeStamp))
+                {
+                    return (DateTime)(NpgsqlTimeStamp)value;
+                }
+                else if (currentType == typeof(NpgsqlTimeStampTZ))
+                {
+                    return (DateTime)(NpgsqlTimeStampTZ)value;
+                }
+                else if (currentType == typeof(NpgsqlInterval))
+                {
+                    return new DateTime(((TimeSpan)(NpgsqlInterval)value).Ticks);
+                }
+#if NET35
+                else if (currentType == typeof(DateTimeOffset))
+                {
+                    return ((DateTimeOffset)value).LocalDateTime;
+                }
+#endif
+                else if (currentType == typeof(TimeSpan))
+                {
+                    return new DateTime(((TimeSpan)value).Ticks);
+                }
+                else
+                {
+                    return DateTime.Parse(value.ToString(), CultureInfo.InvariantCulture);
+                }
+            }
+            else if (expectedType == typeof(byte[]))
+            {
+                if (currentType == typeof(Guid))
+                {
+                    return ((Guid)value).ToByteArray();
+                }
+                else if (value is Array)
+                {
+                    Array valueArray = (Array)value;
+                    int byteLength = Buffer.ByteLength(valueArray);
+                    byte[] bytes = new byte[byteLength];
+                    Buffer.BlockCopy(valueArray, 0, bytes, 0, byteLength);
+                    return bytes;
+                }
+                else
+                {
+                    // expect InvalidCastException from this call
+                    return Convert.ChangeType(value, expectedType);
+                }
+            }
+            else // long, int, short, double, float, decimal, byte, sbyte, bool, and other unspecified types
+            {
+                // ChangeType supports the conversions we want for above expected types
+                return Convert.ChangeType(value, expectedType);
+            }
+        }
+
+	}
 }
